@@ -1,7 +1,6 @@
-/* SPDX-License-Identifier: BSD-2-Clause */
 /*
  * dhcpcd: BPF arp and bootp filtering
- * Copyright (c) 2006-2021 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2018 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -57,6 +56,8 @@
 #include "if.h"
 #include "logerr.h"
 
+#define	ARP_ADDRS_MAX	3
+
 /* BPF helper macros */
 #ifdef __linux__
 #define	BPF_WHOLEPACKET		0x7fffffff /* work around buggy LPF filters */
@@ -70,20 +71,20 @@
 	(insn)->jt = 0;						\
 	(insn)->jf = 0;						\
 	(insn)->k = (uint32_t)(v);				\
-}
+};
 
 #define	BPF_SET_JUMP(insn, c, v, t, f) {			\
 	(insn)->code = (c);					\
 	(insn)->jt = (t);					\
 	(insn)->jf = (f);					\
 	(insn)->k = (uint32_t)(v);				\
-}
+};
 
 size_t
 bpf_frame_header_len(const struct interface *ifp)
 {
 
-	switch (ifp->hwtype) {
+	switch(ifp->family) {
 	case ARPHRD_ETHER:
 		return sizeof(struct ether_header);
 	default:
@@ -91,84 +92,30 @@ bpf_frame_header_len(const struct interface *ifp)
 	}
 }
 
-void *
-bpf_frame_header_src(const struct interface *ifp, void *fh, size_t *len)
-{
-	uint8_t *f = fh;
-
-	switch (ifp->hwtype) {
-	case ARPHRD_ETHER:
-		*len = sizeof(((struct ether_header *)0)->ether_shost);
-		return f + offsetof(struct ether_header, ether_shost);
-	default:
-		*len = 0;
-		errno =	ENOTSUP;
-		return NULL;
-	}
-}
-
-void *
-bpf_frame_header_dst(const struct interface *ifp, void *fh, size_t *len)
-{
-	uint8_t *f = fh;
-
-	switch (ifp->hwtype) {
-	case ARPHRD_ETHER:
-		*len = sizeof(((struct ether_header *)0)->ether_dhost);
-		return f + offsetof(struct ether_header, ether_dhost);
-	default:
-		*len = 0;
-		errno =	ENOTSUP;
-		return NULL;
-	}
-}
-
-static const uint8_t etherbcastaddr[] =
-    { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
-
-int
-bpf_frame_bcast(const struct interface *ifp, const void *frame)
-{
-
-	switch (ifp->hwtype) {
-	case ARPHRD_ETHER:
-		return memcmp((const char *)frame +
-		    offsetof(struct ether_header, ether_dhost),
-		    etherbcastaddr, sizeof(etherbcastaddr));
-	default:
-		return -1;
-	}
-}
-
 #ifndef __linux__
 /* Linux is a special snowflake for opening, attaching and reading BPF.
  * See if-linux.c for the Linux specific BPF functions. */
 
 const char *bpf_name = "Berkley Packet Filter";
 
-struct bpf *
-bpf_open(const struct interface *ifp,
-    int (*filter)(const struct bpf *, const struct in_addr *),
-    const struct in_addr *ia)
+int
+bpf_open(struct interface *ifp, int (*filter)(struct interface *, int))
 {
-	struct bpf *bpf;
-	struct bpf_version pv = { .bv_major = 0, .bv_minor = 0 };
-	struct ifreq ifr = { .ifr_flags = 0 };
+	struct ipv4_state *state;
+	int fd = -1;
+	struct ifreq ifr;
 	int ibuf_len = 0;
+	size_t buf_len;
+	struct bpf_version pv;
 #ifdef BIOCIMMEDIATE
-	unsigned int flags;
+	int flags;
 #endif
 #ifndef O_CLOEXEC
 	int fd_opts;
 #endif
 
-	bpf = calloc(1, sizeof(*bpf));
-	if (bpf == NULL)
-		return NULL;
-	bpf->bpf_ifp = ifp;
-
 #ifdef _PATH_BPF
-	bpf->bpf_fd = open(_PATH_BPF, O_RDWR | O_NONBLOCK
+	fd = open(_PATH_BPF, O_RDWR | O_NONBLOCK
 #ifdef O_CLOEXEC
 		| O_CLOEXEC
 #endif
@@ -179,24 +126,27 @@ bpf_open(const struct interface *ifp,
 
 	do {
 		snprintf(device, sizeof(device), "/dev/bpf%d", n++);
-		bpf->bpf_fd = open(device, O_RDWR | O_NONBLOCK
+		fd = open(device, O_RDWR | O_NONBLOCK
 #ifdef O_CLOEXEC
 				| O_CLOEXEC
 #endif
 		);
-	} while (bpf->bpf_fd == -1 && errno == EBUSY);
+	} while (fd == -1 && errno == EBUSY);
 #endif
 
-	if (bpf->bpf_fd == -1)
-		goto eexit;
+	if (fd == -1)
+		return -1;
 
 #ifndef O_CLOEXEC
-	if ((fd_opts = fcntl(bpf->bpf_fd, F_GETFD)) == -1 ||
-	    fcntl(bpf->bpf_fd, F_SETFD, fd_opts | FD_CLOEXEC) == -1)
-		goto eexit;
+	if ((fd_opts = fcntl(fd, F_GETFD)) == -1 ||
+	    fcntl(fd, F_SETFD, fd_opts | FD_CLOEXEC) == -1) {
+		close(fd);
+		return -1;
+	}
 #endif
 
-	if (ioctl(bpf->bpf_fd, BIOCVERSION, &pv) == -1)
+	memset(&pv, 0, sizeof(pv));
+	if (ioctl(fd, BIOCVERSION, &pv) == -1)
 		goto eexit;
 	if (pv.bv_major != BPF_MAJOR_VERSION ||
 	    pv.bv_minor < BPF_MINOR_VERSION) {
@@ -204,84 +154,91 @@ bpf_open(const struct interface *ifp,
 		goto eexit;
 	}
 
+	if (filter(ifp, fd) != 0)
+		goto eexit;
+
+	memset(&ifr, 0, sizeof(ifr));
 	strlcpy(ifr.ifr_name, ifp->name, sizeof(ifr.ifr_name));
-	if (ioctl(bpf->bpf_fd, BIOCSETIF, &ifr) == -1)
-		goto eexit;
-
-#ifdef BIOCIMMEDIATE
-	flags = 1;
-	if (ioctl(bpf->bpf_fd, BIOCIMMEDIATE, &flags) == -1)
-		goto eexit;
-#endif
-
-	if (filter(bpf, ia) != 0)
+	if (ioctl(fd, BIOCSETIF, &ifr) == -1)
 		goto eexit;
 
 	/* Get the required BPF buffer length from the kernel. */
-	if (ioctl(bpf->bpf_fd, BIOCGBLEN, &ibuf_len) == -1)
+	if (ioctl(fd, BIOCGBLEN, &ibuf_len) == -1)
 		goto eexit;
-	bpf->bpf_size = (size_t)ibuf_len;
-	bpf->bpf_buffer = malloc(bpf->bpf_size);
-	if (bpf->bpf_buffer == NULL)
+	buf_len = (size_t)ibuf_len;
+	state = ipv4_getstate(ifp);
+	if (state == NULL)
 		goto eexit;
-	return bpf;
+	if (state->buffer_size != buf_len) {
+		void *nb;
+
+		if ((nb = realloc(state->buffer, buf_len)) == NULL)
+			goto eexit;
+		state->buffer = nb;
+		state->buffer_size = buf_len;
+	}
+
+#ifdef BIOCIMMEDIATE
+	flags = 1;
+	if (ioctl(fd, BIOCIMMEDIATE, &flags) == -1)
+		goto eexit;
+#endif
+
+	return fd;
 
 eexit:
-	if (bpf->bpf_fd != -1)
-		close(bpf->bpf_fd);
-	free(bpf);
-	return NULL;
+	close(fd);
+	return -1;
 }
 
 /* BPF requires that we read the entire buffer.
  * So we pass the buffer in the API so we can loop on >1 packet. */
 ssize_t
-bpf_read(struct bpf *bpf, void *data, size_t len)
+bpf_read(struct interface *ifp, int fd, void *data, size_t len,
+    unsigned int *flags)
 {
+	ssize_t fl = (ssize_t)bpf_frame_header_len(ifp);
 	ssize_t bytes;
+	struct ipv4_state *state = IPV4_STATE(ifp);
+
 	struct bpf_hdr packet;
 	const char *payload;
 
-	bpf->bpf_flags &= ~BPF_EOF;
+	*flags &= ~BPF_EOF;
 	for (;;) {
-		if (bpf->bpf_len == 0) {
-			bytes = read(bpf->bpf_fd, bpf->bpf_buffer,
-			    bpf->bpf_size);
+		if (state->buffer_len == 0) {
+			bytes = read(fd, state->buffer, state->buffer_size);
 #if defined(__sun)
 			/* After 2^31 bytes, the kernel offset overflows.
 			 * To work around this bug, lseek 0. */
 			if (bytes == -1 && errno == EINVAL) {
-				lseek(bpf->bpf_fd, 0, SEEK_SET);
+				lseek(fd, 0, SEEK_SET);
 				continue;
 			}
 #endif
 			if (bytes == -1 || bytes == 0)
 				return bytes;
-			bpf->bpf_len = (size_t)bytes;
-			bpf->bpf_pos = 0;
+			state->buffer_len = (size_t)bytes;
+			state->buffer_pos = 0;
 		}
 		bytes = -1;
-		payload = (const char *)bpf->bpf_buffer + bpf->bpf_pos;
-		memcpy(&packet, payload, sizeof(packet));
-		if (bpf->bpf_pos + packet.bh_caplen + packet.bh_hdrlen >
-		    bpf->bpf_len)
+		memcpy(&packet, state->buffer + state->buffer_pos,
+		    sizeof(packet));
+		if (state->buffer_pos + packet.bh_caplen + packet.bh_hdrlen >
+		    state->buffer_len)
 			goto next; /* Packet beyond buffer, drop. */
-		payload += packet.bh_hdrlen;
-		if (packet.bh_caplen > len)
+		payload = state->buffer + state->buffer_pos +
+		    packet.bh_hdrlen + fl;
+		bytes = (ssize_t)packet.bh_caplen - fl;
+		if ((size_t)bytes > len)
 			bytes = (ssize_t)len;
-		else
-			bytes = (ssize_t)packet.bh_caplen;
-		if (bpf_frame_bcast(bpf->bpf_ifp, payload) == 0)
-			bpf->bpf_flags |= BPF_BCAST;
-		else
-			bpf->bpf_flags &= ~BPF_BCAST;
 		memcpy(data, payload, (size_t)bytes);
 next:
-		bpf->bpf_pos += BPF_WORDALIGN(packet.bh_hdrlen +
+		state->buffer_pos += BPF_WORDALIGN(packet.bh_hdrlen +
 		    packet.bh_caplen);
-		if (bpf->bpf_pos >= bpf->bpf_len) {
-			bpf->bpf_len = bpf->bpf_pos = 0;
-			bpf->bpf_flags |= BPF_EOF;
+		if (state->buffer_pos >= state->buffer_len) {
+			state->buffer_len = state->buffer_pos = 0;
+			*flags |= BPF_EOF;
 		}
 		if (bytes != -1)
 			return bytes;
@@ -293,38 +250,29 @@ next:
 int
 bpf_attach(int fd, void *filter, unsigned int filter_len)
 {
-	struct bpf_program pf = { .bf_insns = filter, .bf_len = filter_len };
+	struct bpf_program pf;
 
 	/* Install the filter. */
+	memset(&pf, 0, sizeof(pf));
+	pf.bf_insns = filter;
+	pf.bf_len = filter_len;
 	return ioctl(fd, BIOCSETF, &pf);
 }
-
-#ifdef BIOCSETWF
-static int
-bpf_wattach(int fd, void *filter, unsigned int filter_len)
-{
-	struct bpf_program pf = { .bf_insns = filter, .bf_len = filter_len };
-
-	/* Install the filter. */
-	return ioctl(fd, BIOCSETWF, &pf);
-}
-#endif
 #endif
 
 #ifndef __sun
 /* SunOS is special too - sending via BPF goes nowhere. */
 ssize_t
-bpf_send(const struct bpf *bpf, uint16_t protocol,
+bpf_send(const struct interface *ifp, int fd, uint16_t protocol,
     const void *data, size_t len)
 {
 	struct iovec iov[2];
 	struct ether_header eh;
 
-	switch(bpf->bpf_ifp->hwtype) {
+	switch(ifp->family) {
 	case ARPHRD_ETHER:
 		memset(&eh.ether_dhost, 0xff, sizeof(eh.ether_dhost));
-		memcpy(&eh.ether_shost, bpf->bpf_ifp->hwaddr,
-		    sizeof(eh.ether_shost));
+		memcpy(&eh.ether_shost, ifp->hwaddr, sizeof(eh.ether_shost));
 		eh.ether_type = htons(protocol);
 		iov[0].iov_base = &eh;
 		iov[0].iov_len = sizeof(eh);
@@ -336,24 +284,26 @@ bpf_send(const struct bpf *bpf, uint16_t protocol,
 	}
 	iov[1].iov_base = UNCONST(data);
 	iov[1].iov_len = len;
-	return writev(bpf->bpf_fd, iov, 2);
+	return writev(fd, iov, 2);
 }
 #endif
 
-void
-bpf_close(struct bpf *bpf)
+int
+bpf_close(struct interface *ifp, int fd)
 {
+	struct ipv4_state *state = IPV4_STATE(ifp);
 
-	close(bpf->bpf_fd);
-	free(bpf->bpf_buffer);
-	free(bpf);
+	/* Rewind the buffer on closing. */
+	state->buffer_len = state->buffer_pos = 0;
+	return close(fd);
 }
 
+/* Normally this is needed by bootp.
+ * Once that uses this again, the ARP guard here can be removed. */
 #ifdef ARP
-#define BPF_CMP_HWADDR_LEN	((((HWADDR_LEN / 4) + 2) * 2) + 1)
 static unsigned int
 bpf_cmp_hwaddr(struct bpf_insn *bpf, size_t bpf_len, size_t off,
-    bool equal, const uint8_t *hwaddr, size_t hwaddr_len)
+    bool equal, uint8_t *hwaddr, size_t hwaddr_len)
 {
 	struct bpf_insn *bp;
 	size_t maclen, nlft, njmps;
@@ -437,7 +387,13 @@ bpf_cmp_hwaddr(struct bpf_insn *bpf, size_t bpf_len, size_t off,
 #endif
 
 #ifdef ARP
+
 static const struct bpf_insn bpf_arp_ether [] = {
+	/* Ensure packet is at least correct size. */
+	BPF_STMT(BPF_LD + BPF_W + BPF_LEN, 0),
+	BPF_JUMP(BPF_JMP + BPF_JGE + BPF_K, sizeof(struct ether_arp), 1, 0),
+	BPF_STMT(BPF_RET + BPF_K, 0),
+
 	/* Check this is an ARP packet. */
 	BPF_STMT(BPF_LD + BPF_H + BPF_ABS,
 	         offsetof(struct ether_header, ether_type)),
@@ -447,7 +403,7 @@ static const struct bpf_insn bpf_arp_ether [] = {
 	/* Load frame header length into X */
 	BPF_STMT(BPF_LDX + BPF_W + BPF_IMM, sizeof(struct ether_header)),
 
-	/* Make sure the hardware type matches. */
+	/* Make sure the hardware family matches. */
 	BPF_STMT(BPF_LD + BPF_H + BPF_IND, offsetof(struct arphdr, ar_hrd)),
 	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, ARPHRD_ETHER, 1, 0),
 	BPF_STMT(BPF_RET + BPF_K, 0),
@@ -455,10 +411,10 @@ static const struct bpf_insn bpf_arp_ether [] = {
 	/* Make sure the hardware length matches. */
 	BPF_STMT(BPF_LD + BPF_B + BPF_IND, offsetof(struct arphdr, ar_hln)),
 	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K,
-	         sizeof(((struct ether_arp *)0)->arp_sha), 1, 0),
+	         sizeof((struct ether_arp *)0)->arp_sha, 1, 0),
 	BPF_STMT(BPF_RET + BPF_K, 0),
 };
-#define BPF_ARP_ETHER_LEN	__arraycount(bpf_arp_ether)
+#define bpf_arp_ether_len	__arraycount(bpf_arp_ether)
 
 static const struct bpf_insn bpf_arp_filter [] = {
 	/* Make sure this is for IP. */
@@ -469,36 +425,34 @@ static const struct bpf_insn bpf_arp_filter [] = {
 	BPF_STMT(BPF_LD + BPF_H + BPF_IND, offsetof(struct arphdr, ar_op)),
 	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, ARPOP_REQUEST, 2, 0),
 	/* or ARP REPLY. */
-	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, ARPOP_REPLY, 1, 0),
+	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, ARPOP_REPLY, 1, 1),
 	BPF_STMT(BPF_RET + BPF_K, 0),
 	/* Make sure the protocol length matches. */
 	BPF_STMT(BPF_LD + BPF_B + BPF_IND, offsetof(struct arphdr, ar_pln)),
 	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, sizeof(in_addr_t), 1, 0),
 	BPF_STMT(BPF_RET + BPF_K, 0),
 };
-#define BPF_ARP_FILTER_LEN	__arraycount(bpf_arp_filter)
+#define bpf_arp_filter_len	__arraycount(bpf_arp_filter)
+#define bpf_arp_extra		((((ARP_ADDRS_MAX + 1) * 2) * 2) + 2)
+#define bpf_arp_hw		((((HWADDR_LEN / 4) + 2) * 2) + 1)
 
-/* One address is two checks of two statements. */
-#define BPF_NADDRS		1
-#define BPF_ARP_ADDRS_LEN	5 + ((BPF_NADDRS * 2) * 2)
-
-#define BPF_ARP_LEN		BPF_ARP_ETHER_LEN + BPF_ARP_FILTER_LEN + \
-				BPF_CMP_HWADDR_LEN + BPF_ARP_ADDRS_LEN
-
-static int
-bpf_arp_rw(const struct bpf *bpf, const struct in_addr *ia, bool recv)
+int
+bpf_arp(struct interface *ifp, int fd)
 {
-	const struct interface *ifp = bpf->bpf_ifp;
-	struct bpf_insn buf[BPF_ARP_LEN + 1];
+	struct bpf_insn bpf[3+ bpf_arp_filter_len + bpf_arp_hw + bpf_arp_extra];
 	struct bpf_insn *bp;
+	struct iarp_state *state;
 	uint16_t arp_len;
 
-	bp = buf;
+	if (fd == -1)
+		return 0;
+
+	bp = bpf;
 	/* Check frame header. */
-	switch(ifp->hwtype) {
+	switch(ifp->family) {
 	case ARPHRD_ETHER:
 		memcpy(bp, bpf_arp_ether, sizeof(bpf_arp_ether));
-		bp += BPF_ARP_ETHER_LEN;
+		bp += bpf_arp_ether_len;
 		arp_len = sizeof(struct ether_header)+sizeof(struct ether_arp);
 		break;
 	default:
@@ -508,69 +462,67 @@ bpf_arp_rw(const struct bpf *bpf, const struct in_addr *ia, bool recv)
 
 	/* Copy in the main filter. */
 	memcpy(bp, bpf_arp_filter, sizeof(bpf_arp_filter));
-	bp += BPF_ARP_FILTER_LEN;
+	bp += bpf_arp_filter_len;
 
 	/* Ensure it's not from us. */
-	bp += bpf_cmp_hwaddr(bp, BPF_CMP_HWADDR_LEN, sizeof(struct arphdr),
-	                     !recv, ifp->hwaddr, ifp->hwlen);
+	bp += bpf_cmp_hwaddr(bp, bpf_arp_hw, sizeof(struct arphdr),
+	                     false, ifp->hwaddr, ifp->hwlen);
 
-	/* Match sender protocol address */
-	BPF_SET_STMT(bp, BPF_LD + BPF_W + BPF_IND,
-	    sizeof(struct arphdr) + ifp->hwlen);
-	bp++;
-	BPF_SET_JUMP(bp, BPF_JMP + BPF_JEQ + BPF_K, htonl(ia->s_addr), 0, 1);
-	bp++;
-	BPF_SET_STMT(bp, BPF_RET + BPF_K, arp_len);
-	bp++;
+	state = ARP_STATE(ifp);
+	if (TAILQ_FIRST(&state->arp_states)) {
+		struct arp_state *astate;
+		size_t naddrs;
 
-	/* If we didn't match sender, then we're only interested in
-	 * ARP probes to us, so check the null host sender. */
-	BPF_SET_JUMP(bp, BPF_JMP + BPF_JEQ + BPF_K, INADDR_ANY, 1, 0);
-	bp++;
-	BPF_SET_STMT(bp, BPF_RET + BPF_K, 0);
-	bp++;
+		/* Match sender protocol address */
+		BPF_SET_STMT(bp, BPF_LD + BPF_W + BPF_IND,
+		             sizeof(struct arphdr) + ifp->hwlen);
+		bp++;
+		naddrs = 0;
+		TAILQ_FOREACH(astate, &state->arp_states, next) {
+			if (++naddrs > ARP_ADDRS_MAX) {
+				errno = ENOBUFS;
+				logerr(__func__);
+				break;
+			}
+			BPF_SET_JUMP(bp, BPF_JMP + BPF_JEQ + BPF_K,
+			             htonl(astate->addr.s_addr), 0, 1);
+			bp++;
+			BPF_SET_STMT(bp, BPF_RET + BPF_K, arp_len);
+			bp++;
+		}
 
-	/* Match target protocol address */
-	BPF_SET_STMT(bp, BPF_LD + BPF_W + BPF_IND, (sizeof(struct arphdr) +
-	    (size_t)(ifp->hwlen * 2) + sizeof(in_addr_t)));
-	bp++;
-	BPF_SET_JUMP(bp, BPF_JMP + BPF_JEQ + BPF_K, htonl(ia->s_addr), 0, 1);
-	bp++;
-	BPF_SET_STMT(bp, BPF_RET + BPF_K, arp_len);
-	bp++;
+		/* If we didn't match sender, then we're only interested in
+		 * ARP probes to us, so check the null host sender. */
+		BPF_SET_JUMP(bp, BPF_JMP + BPF_JEQ + BPF_K, INADDR_ANY, 1, 0);
+		bp++;
+		BPF_SET_STMT(bp, BPF_RET + BPF_K, 0);
+		bp++;
 
-	/* No match, drop it */
-	BPF_SET_STMT(bp, BPF_RET + BPF_K, 0);
-	bp++;
+		/* Match target protocol address */
+		BPF_SET_STMT(bp, BPF_LD + BPF_W + BPF_IND,
+		             (sizeof(struct arphdr)
+			     + (size_t)(ifp->hwlen * 2) + sizeof(in_addr_t)));
+		bp++;
+		naddrs = 0;
+		TAILQ_FOREACH(astate, &state->arp_states, next) {
+			if (++naddrs > ARP_ADDRS_MAX) {
+				/* Already logged error above. */
+				break;
+			}
+			BPF_SET_JUMP(bp, BPF_JMP + BPF_JEQ + BPF_K,
+			             htonl(astate->addr.s_addr), 0, 1);
+			bp++;
+			BPF_SET_STMT(bp, BPF_RET + BPF_K, arp_len);
+			bp++;
+		}
 
-#ifdef BIOCSETWF
-	if (!recv)
-		return bpf_wattach(bpf->bpf_fd, buf, (unsigned int)(bp - buf));
-#endif
+		/* Return nothing, no protocol address match. */
+		BPF_SET_STMT(bp, BPF_RET + BPF_K, 0);
+		bp++;
+	}
 
-	return bpf_attach(bpf->bpf_fd, buf, (unsigned int)(bp - buf));
+	return bpf_attach(fd, bpf, (unsigned int)(bp - bpf));
 }
-
-int
-bpf_arp(const struct bpf *bpf, const struct in_addr *ia)
-{
-
-#ifdef BIOCSETWF
-	if (bpf_arp_rw(bpf, ia, true) == -1 ||
-	    bpf_arp_rw(bpf, ia, false) == -1 ||
-	    ioctl(bpf->bpf_fd, BIOCLOCK) == -1)
-		return -1;
-	return 0;
-#else
-	return bpf_arp_rw(bpf, ia, true);
-#endif
-}
-#endif
-
-#ifdef ARPHRD_NONE
-static const struct bpf_insn bpf_bootp_none[] = {
-};
-#define BPF_BOOTP_NONE_LEN	__arraycount(bpf_bootp_none)
 #endif
 
 static const struct bpf_insn bpf_bootp_ether[] = {
@@ -580,16 +532,17 @@ static const struct bpf_insn bpf_bootp_ether[] = {
 	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, ETHERTYPE_IP, 1, 0),
 	BPF_STMT(BPF_RET + BPF_K, 0),
 
-	/* Advance to the IP header. */
-	BPF_STMT(BPF_LDX + BPF_K, sizeof(struct ether_header)),
+	/* Load frame header length into X. */
+	BPF_STMT(BPF_LDX + BPF_W + BPF_IMM, sizeof(struct ether_header)),
+	/* Copy to M0. */
+	BPF_STMT(BPF_STX, 0),
 };
 #define BPF_BOOTP_ETHER_LEN	__arraycount(bpf_bootp_ether)
 
-static const struct bpf_insn bpf_bootp_base[] = {
+static const struct bpf_insn bpf_bootp_filter[] = {
 	/* Make sure it's an IPv4 packet. */
 	BPF_STMT(BPF_LD + BPF_B + BPF_IND, 0),
-	BPF_STMT(BPF_ALU + BPF_AND + BPF_K, 0xf0),
-	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, 0x40, 1, 0),
+	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, 0x45, 1, 0),
 	BPF_STMT(BPF_RET + BPF_K, 0),
 
 	/* Make sure it's a UDP packet. */
@@ -602,55 +555,70 @@ static const struct bpf_insn bpf_bootp_base[] = {
 	BPF_JUMP(BPF_JMP + BPF_JSET + BPF_K, 0x1fff, 0, 1),
 	BPF_STMT(BPF_RET + BPF_K, 0),
 
+	/* Store IP location in M1. */
+	BPF_STMT(BPF_LD + BPF_H + BPF_IND, offsetof(struct ip, ip_len)),
+	BPF_STMT(BPF_ST, 1),
+
+	/* Store IP length in M2. */
+	BPF_STMT(BPF_LD + BPF_H + BPF_IND, offsetof(struct ip, ip_len)),
+	BPF_STMT(BPF_ST, 2),
+
 	/* Advance to the UDP header. */
-	BPF_STMT(BPF_LD + BPF_B + BPF_IND, 0),
-	BPF_STMT(BPF_ALU + BPF_AND + BPF_K, 0x0f),
-	BPF_STMT(BPF_ALU + BPF_MUL + BPF_K, 4),
-	BPF_STMT(BPF_ALU + BPF_ADD + BPF_X, 0),
+	BPF_STMT(BPF_MISC + BPF_TXA, 0),
+	BPF_STMT(BPF_ALU + BPF_ADD + BPF_K, sizeof(struct ip)),
 	BPF_STMT(BPF_MISC + BPF_TAX, 0),
-};
-#define BPF_BOOTP_BASE_LEN	__arraycount(bpf_bootp_base)
+
+	/* Store X in M3. */
+	BPF_STMT(BPF_STX, 3),
 
-static const struct bpf_insn bpf_bootp_read[] = {
 	/* Make sure it's from and to the right port. */
 	BPF_STMT(BPF_LD + BPF_W + BPF_IND, 0),
 	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, (BOOTPS << 16) + BOOTPC, 1, 0),
 	BPF_STMT(BPF_RET + BPF_K, 0),
-};
-#define BPF_BOOTP_READ_LEN	__arraycount(bpf_bootp_read)
 
-#ifdef BIOCSETWF
-static const struct bpf_insn bpf_bootp_write[] = {
-	/* Make sure it's from and to the right port. */
-	BPF_STMT(BPF_LD + BPF_W + BPF_IND, 0),
-	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, (BOOTPC << 16) + BOOTPS, 1, 0),
+	/* Store UDP length in X. */
+	BPF_STMT(BPF_LD + BPF_H + BPF_IND, offsetof(struct udphdr, uh_ulen)),
+	BPF_STMT(BPF_MISC + BPF_TAX, 0),
+	/* Copy IP length in M2 to A. */
+	BPF_STMT(BPF_LD + BPF_MEM, 2),
+	/* Ensure IP length - IP header size == UDP length. */
+	BPF_STMT(BPF_ALU + BPF_SUB + BPF_K, sizeof(struct ip)),
+	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_X, 0, 1, 0),
+	BPF_STMT(BPF_RET + BPF_K, 0),
+
+	/* Advance to the BOOTP packet (UDP X is in M3). */
+	BPF_STMT(BPF_LD + BPF_MEM, 3),
+	BPF_STMT(BPF_ALU + BPF_ADD + BPF_K, sizeof(struct udphdr)),
+	BPF_STMT(BPF_MISC + BPF_TAX, 0),
+
+	/* Make sure it's BOOTREPLY. */
+	BPF_STMT(BPF_LD + BPF_B + BPF_IND, offsetof(struct bootp, op)),
+	BPF_JUMP(BPF_JMP + BPF_JEQ + BPF_K, BOOTREPLY, 1, 0),
 	BPF_STMT(BPF_RET + BPF_K, 0),
 };
-#define BPF_BOOTP_WRITE_LEN	__arraycount(bpf_bootp_write)
-#endif
 
+#define BPF_BOOTP_FILTER_LEN	__arraycount(bpf_bootp_filter)
 #define BPF_BOOTP_CHADDR_LEN	((BOOTP_CHADDR_LEN / 4) * 3)
 #define	BPF_BOOTP_XID_LEN	4 /* BOUND check is 4 instructions */
 
-#define BPF_BOOTP_LEN		BPF_BOOTP_ETHER_LEN + \
-				BPF_BOOTP_BASE_LEN + BPF_BOOTP_READ_LEN + \
-				BPF_BOOTP_XID_LEN + BPF_BOOTP_CHADDR_LEN + 4
+#define BPF_BOOTP_LEN		BPF_BOOTP_ETHER_LEN + BPF_BOOTP_FILTER_LEN \
+				+ BPF_BOOTP_XID_LEN + BPF_BOOTP_CHADDR_LEN + 4
 
-static int
-bpf_bootp_rw(const struct bpf *bpf, bool read)
+int
+bpf_bootp(struct interface *ifp, int fd)
 {
-	struct bpf_insn buf[BPF_BOOTP_LEN + 1];
+#if 0
+	const struct dhcp_state *state = D_CSTATE(ifp);
+#endif
+	struct bpf_insn bpf[BPF_BOOTP_LEN];
 	struct bpf_insn *bp;
 
-	bp = buf;
+	if (fd == -1)
+		return 0;
+
+	bp = bpf;
 	/* Check frame header. */
-	switch(bpf->bpf_ifp->hwtype) {
-#ifdef ARPHRD_NONE
-	case ARPHRD_NONE:
-		memcpy(bp, bpf_bootp_none, sizeof(bpf_bootp_none));
-		bp += BPF_BOOTP_NONE_LEN;
-		break;
-#endif
+	switch(ifp->family) {
 	case ARPHRD_ETHER:
 		memcpy(bp, bpf_bootp_ether, sizeof(bpf_bootp_ether));
 		bp += BPF_BOOTP_ETHER_LEN;
@@ -661,53 +629,55 @@ bpf_bootp_rw(const struct bpf *bpf, bool read)
 	}
 
 	/* Copy in the main filter. */
-	memcpy(bp, bpf_bootp_base, sizeof(bpf_bootp_base));
-	bp += BPF_BOOTP_BASE_LEN;
+	memcpy(bp, bpf_bootp_filter, sizeof(bpf_bootp_filter));
+	bp += BPF_BOOTP_FILTER_LEN;
 
-#ifdef BIOCSETWF
-	if (!read) {
-		memcpy(bp, bpf_bootp_write, sizeof(bpf_bootp_write));
-		bp += BPF_BOOTP_WRITE_LEN;
+	/* These checks won't work when same IP exists on other interfaces. */
+#if 0
+	if (ifp->hwlen <= sizeof(((struct bootp *)0)->chaddr))
+		bp += bpf_cmp_hwaddr(bp, BPF_BOOTP_CHADDR_LEN,
+		                     offsetof(struct bootp, chaddr),
+				     true, ifp->hwaddr, ifp->hwlen);
 
-		/* All passed, return the packet. */
-		BPF_SET_STMT(bp, BPF_RET + BPF_K, BPF_WHOLEPACKET);
+	/* Make sure the BOOTP packet is for us. */
+	if (state->state == DHS_BOUND) {
+		/* If bound, we only expect FORCERENEW messages
+		 * and they need to be unicast to us.
+		 * Move back to the IP header in M0 and check dst. */
+		BPF_SET_STMT(bp, BPF_LDX + BPF_W + BPF_MEM, 0);
+		bp++;
+		BPF_SET_STMT(bp, BPF_LD + BPF_W + BPF_IND,
+		             offsetof(struct ip, ip_dst));
+		bp++;
+		BPF_SET_JUMP(bp, BPF_JMP + BPF_JEQ + BPF_K,
+		             htonl(state->lease.addr.s_addr), 1, 0);
+		bp++;
+		BPF_SET_STMT(bp, BPF_RET + BPF_K, 0);
+		bp++;
+	} else {
+		/* As we're not bound, we need to check xid to ensure
+		 * it's a reply to our transaction. */
+		BPF_SET_STMT(bp, BPF_LD + BPF_W + BPF_IND,
+		             offsetof(struct bootp, xid));
+		bp++;
+		BPF_SET_JUMP(bp, BPF_JMP + BPF_JEQ + BPF_K,
+		             state->xid, 1, 0);
+		bp++;
+		BPF_SET_STMT(bp, BPF_RET + BPF_K, 0);
 		bp++;
-
-		return bpf_wattach(bpf->bpf_fd, buf, (unsigned int)(bp - buf));
 	}
-#else
-	UNUSED(read);
 #endif
 
-	memcpy(bp, bpf_bootp_read, sizeof(bpf_bootp_read));
-	bp += BPF_BOOTP_READ_LEN;
-
-	/* All passed, return the packet. */
-	BPF_SET_STMT(bp, BPF_RET + BPF_K, BPF_WHOLEPACKET);
+	/* All passed, return the packet
+	 * (Frame length in M0, IP length in M2). */
+	BPF_SET_STMT(bp, BPF_LD + BPF_MEM, 0);
+	bp++;
+	BPF_SET_STMT(bp, BPF_LDX + BPF_MEM, 2);
+	bp++;
+	BPF_SET_STMT(bp, BPF_ALU + BPF_ADD + BPF_X, 0);
+	bp++;
+	BPF_SET_STMT(bp, BPF_RET + BPF_A, 0);
 	bp++;
 
-	return bpf_attach(bpf->bpf_fd, buf, (unsigned int)(bp - buf));
-}
-
-int
-bpf_bootp(const struct bpf *bpf, __unused const struct in_addr *ia)
-{
-
-#ifdef BIOCSETWF
-	if (bpf_bootp_rw(bpf, true) == -1 ||
-	    bpf_bootp_rw(bpf, false) == -1 ||
-	    ioctl(bpf->bpf_fd, BIOCLOCK) == -1)
-		return -1;
-	return 0;
-#else
-#ifdef PRIVSEP
-#if defined(__sun) /* Solaris cannot send via BPF. */
-#elif defined(BIOCSETF)
-#warning No BIOCSETWF support - a compromised BPF can be used as a raw socket
-#else
-#warning A compromised PF_PACKET socket can be used as a raw socket
-#endif
-#endif
-	return bpf_bootp_rw(bpf, true);
-#endif
+	return bpf_attach(fd, bpf, (unsigned int)(bp - bpf));
 }
