@@ -1,7 +1,6 @@
-/* SPDX-License-Identifier: BSD-2-Clause */
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2019 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2018 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -31,9 +30,12 @@
 
 #include <sys/param.h>
 #include <sys/time.h>
-#include <stdint.h>
 #include <stdio.h>
 
+#include "config.h"
+#include "defs.h"
+#include "dhcpcd.h"
+
 #ifndef HOSTNAME_MAX_LEN
 #define HOSTNAME_MAX_LEN	250	/* 255 - 3 (FQDN) - 2 (DNS enc) */
 #endif
@@ -116,6 +118,12 @@
 # ifndef __packed
 #  define __packed __attribute__((__packed__))
 # endif
+# ifndef __sysloglike
+#  ifndef __syslog_attribute_
+#    define __syslog__ __printf__
+#  endif
+#  define __sysloglike(a, b) __attribute__((format(__syslog__, a, b)))
+# endif
 # ifndef __unused
 #  define __unused __attribute__((__unused__))
 # endif
@@ -123,42 +131,14 @@
 # ifndef __packed
 #  define __packed
 # endif
+# ifndef __sysloglike
+#  define __sysloglike
+# endif
 # ifndef __unused
 #  define __unused
 # endif
 #endif
 
-/* Needed for rbtree(3) compat */
-#ifndef __RCSID
-#define __RCSID(a)
-#endif
-#ifndef __predict_false
-# if __GNUC__ > 2
-#  define	__predict_true(exp)	__builtin_expect((exp) != 0, 1)
-#  define	__predict_false(exp)	__builtin_expect((exp) != 0, 0)
-#else
-#  define	__predict_true(exp)	(exp)
-#  define	__predict_false(exp)	(exp)
-# endif
-#endif
-#ifndef __BEGIN_DECLS
-# if defined(__cplusplus)
-#  define	__BEGIN_DECLS		extern "C" {
-#  define	__END_DECLS		};
-# else /* __BEGIN_DECLS */
-#  define	__BEGIN_DECLS
-#  define	__END_DECLS
-# endif /* __BEGIN_DECLS */
-#endif /* __BEGIN_DECLS */
-
-#ifndef __fallthrough
-# if __GNUC__ >= 7
-#  define __fallthrough __attribute__((fallthrough))
-# else
-#  define __fallthrough
-# endif
-#endif
-
 /*
  * Compile Time Assertion.
  */
@@ -170,7 +150,7 @@
 #  define	__CTASSERT99(x, a, b)	__CTASSERT0(x, __CONCAT(__ctassert,a), \
 					       __CONCAT(_,b))
 # endif
-# define	__CTASSERT0(x, y, z)	__CTASSERT1(x, y, z)
+# define	__CTASSERT0(x, y, z)	__CTASSERT1(x, y, z) 
 # define	__CTASSERT1(x, y, z)	typedef char y ## z[/*CONSTCOND*/(x) ? 1 : -1] __unused
 #endif
 
@@ -194,8 +174,14 @@ void get_line_free(void);
 extern int clock_monotonic;
 int get_monotonic(struct timespec *);
 
+ssize_t setvar(char **, const char *, const char *, const char *);
+ssize_t setvard(char **, const char *, const char *, size_t);
+ssize_t addvar(char ***, const char *, const char *, const char *);
+ssize_t addvard(char ***, const char *, const char *, size_t);
+
 const char *hwaddr_ntoa(const void *, size_t, char *, size_t);
 size_t hwaddr_aton(uint8_t *, const char *);
 size_t read_hwaddr_aton(uint8_t **, const char *);
-int is_root_local(void);
+
+ssize_t recvmsg_realloc(int, struct msghdr *, int);
 #endif
