@@ -98,11 +98,6 @@ struct dhcp6_ia_na {
 };
 __CTASSERT(sizeof(struct dhcp6_ia_na) == 12);
 
-struct dhcp6_ia_ta {
-	uint8_t iaid[4];
-};
-__CTASSERT(sizeof(struct dhcp6_ia_ta) == 4);
-
 struct dhcp6_ia_addr {
 	struct in6_addr addr;
 	uint32_t pltime;
@@ -198,48 +193,6 @@ dhcp6_printoptions(const struct dhcpcd_ctx *ctx,
 	}
 }
 
-static size_t
-dhcp6_makeuser(void *data, const struct interface *ifp)
-{
-	const struct if_options *ifo = ifp->options;
-	struct dhcp6_option o;
-	uint8_t *p;
-	const uint8_t *up, *ue;
-	uint16_t ulen, unlen;
-	size_t olen;
-
-	/* Convert the DHCPv4 user class option to DHCPv6 */
-	up = ifo->userclass;
-	ulen = *up++;
-	if (ulen == 0)
-		return 0;
-
-	p = data;
-	olen = 0;
-	if (p != NULL)
-		p += sizeof(o);
-
-	ue = up + ulen;
-	for (; up < ue; up += ulen) {
-		ulen = *up++;
-		olen += sizeof(ulen) + ulen;
-		if (data == NULL)
-			continue;
-		unlen = htons(ulen);
-		memcpy(p, &unlen, sizeof(unlen));
-		p += sizeof(unlen);
-		memcpy(p, up, ulen);
-		p += ulen;
-	}
-	if (data != NULL) {
-		o.code = htons(D6_OPTION_USER_CLASS);
-		o.len = htons((uint16_t)olen);
-		memcpy(data, &o, sizeof(o));
-	}
-
-	return sizeof(o) + olen;
-}
-
 static size_t
 dhcp6_makevendor(void *data, const struct interface *ifp)
 {
@@ -292,7 +245,7 @@ dhcp6_makevendor(void *data, const struct interface *ifp)
 			{
 				hvlen = htons((uint16_t)vivco->len);
 				memcpy(p, &hvlen, sizeof(hvlen));
-				p += sizeof(hvlen);
+				p += sizeof(len);
 				memcpy(p, vivco->data, vivco->len);
 				p += vivco->len;
 			}
@@ -484,18 +437,20 @@ again:
 
 #ifndef SMALL
 static const struct if_sla *
-dhcp6_findselfsla(struct interface *ifp)
+dhcp6_findselfsla(struct interface *ifp, const uint8_t *iaid)
 {
 	size_t i, j;
-	struct if_ia *ia;
 
 	for (i = 0; i < ifp->options->ia_len; i++) {
-		ia = &ifp->options->ia[i];
-		if (ia->ia_type != D6_OPTION_IA_PD)
-			continue;
-		for (j = 0; j < ia->sla_len; j++) {
-			if (strcmp(ia->sla[j].ifname, ifp->name) == 0)
-				return &ia->sla[j];
+		if (iaid == NULL ||
+		    memcmp(&ifp->options->ia[i].iaid, iaid,
+		    sizeof(ifp->options->ia[i].iaid)) == 0)
+		{
+			for (j = 0; j < ifp->options->ia[i].sla_len; j++) {
+				if (strcmp(ifp->options->ia[i].sla[j].ifname,
+				    ifp->name) == 0)
+					return &ifp->options->ia[i].sla[j];
+			}
 		}
 	}
 	return NULL;
@@ -632,7 +587,6 @@ dhcp6_makemessage(struct interface *ifp)
 	int fqdn;
 	struct dhcp6_ia_na ia_na;
 	uint16_t ia_na_len;
-	struct if_ia *ifia;
 #ifdef AUTH
 	uint16_t auth_len;
 #endif
@@ -697,7 +651,7 @@ dhcp6_makemessage(struct interface *ifp)
 				len += sizeof(o.len);
 			}
 		}
-		if (dhcp6_findselfsla(ifp)) {
+		if (dhcp6_findselfsla(ifp, NULL)) {
 			n_options++;
 			len += sizeof(o.len);
 		}
@@ -723,8 +677,6 @@ dhcp6_makemessage(struct interface *ifp)
 	len += sizeof(*state->send);
 	len += sizeof(o) + ifp->ctx->duid_len;
 	len += sizeof(o) + sizeof(uint16_t); /* elapsed */
-	if (!has_option_mask(ifo->nomask6, D6_OPTION_USER_CLASS))
-		len += dhcp6_makeuser(NULL, ifp);
 	if (!has_option_mask(ifo->nomask6, D6_OPTION_VENDOR_CLASS))
 		len += dhcp6_makevendor(NULL, ifp);
 
@@ -776,12 +728,7 @@ dhcp6_makemessage(struct interface *ifp)
 		}
 		/* FALLTHROUGH */
 	case DH6S_INIT:
-		for (l = 0; l < ifo->ia_len; l++) {
-			len += sizeof(o) + sizeof(uint32_t); /* IAID */
-			/* IA_TA does not have T1 or T2 timers */
-			if (ifo->ia[l].ia_type != D6_OPTION_IA_TA)
-				len += sizeof(uint32_t) + sizeof(uint32_t);
-		}
+		len += ifo->ia_len * (sizeof(o) + (sizeof(uint32_t) * 3));
 		IA = 1;
 		break;
 	default:
@@ -830,15 +777,6 @@ dhcp6_makemessage(struct interface *ifp)
 		return -1;
 	}
 
-	/* In non master mode we listen and send from fixed addresses.
-	 * We should try and match an address we have to unicast to,
-	 * but for now this is the safest policy. */
-	if (unicast != NULL && !(ifp->ctx->options & DHCPCD_MASTER)) {
-		logdebugx("%s: ignoring unicast option as not master",
-		    ifp->name);
-		unicast = NULL;
-	}
-
 #ifdef AUTH
 	auth_len = 0;
 	if (ifo->auth.options & DHCPCD_AUTH_SEND) {
@@ -897,8 +835,6 @@ dhcp6_makemessage(struct interface *ifp)
 	si_len = 0;
 	COPYIN(D6_OPTION_ELAPSED, &si_len, sizeof(si_len));
 
-	if (!has_option_mask(ifo->nomask6, D6_OPTION_USER_CLASS))
-		p += dhcp6_makeuser(p, ifp);
 	if (!has_option_mask(ifo->nomask6, D6_OPTION_VENDOR_CLASS))
 		p += dhcp6_makevendor(p, ifp);
 
@@ -908,29 +844,19 @@ dhcp6_makemessage(struct interface *ifp)
 		COPYIN1(D6_OPTION_RAPID_COMMIT, 0);
 
 	for (l = 0; IA && l < ifo->ia_len; l++) {
-		ifia = &ifo->ia[l];
 		o_lenp = NEXTLEN;
-		/* TA structure is the same as the others,
-		 * it just lacks the T1 and T2 timers.
-		 * These happen to be at the end of the struct,
-		 * so we just don't copy them in. */
-		if (ifia->ia_type == D6_OPTION_IA_TA)
-			ia_na_len = sizeof(struct dhcp6_ia_ta);
-		else
-			ia_na_len = sizeof(ia_na);
-		memcpy(ia_na.iaid, ifia->iaid, sizeof(ia_na.iaid));
+		ia_na_len = sizeof(ia_na);
+		memcpy(ia_na.iaid, ifo->ia[l].iaid, sizeof(ia_na.iaid));
 		ia_na.t1 = 0;
 		ia_na.t2 = 0;
-		COPYIN(ifia->ia_type, &ia_na, ia_na_len);
+		COPYIN(ifo->ia[l].ia_type, &ia_na, sizeof(ia_na));
 		TAILQ_FOREACH(ap, &state->addrs, next) {
 			if (ap->flags & IPV6_AF_STALE)
 				continue;
 			if (ap->prefix_vltime == 0 &&
 			    !(ap->flags & IPV6_AF_REQUEST))
 				continue;
-			if (ap->ia_type != ifia->ia_type)
-				continue;
-			if (memcmp(ap->iaid, ifia->iaid, sizeof(ap->iaid)))
+			if (memcmp(ifo->ia[l].iaid, ap->iaid, sizeof(uint32_t)))
 				continue;
 			if (ap->ia_type == D6_OPTION_IA_PD) {
 #ifndef SMALL
@@ -1073,7 +999,7 @@ dhcp6_makemessage(struct interface *ifp)
 					    (o.len + sizeof(o.code));
 				}
 			}
-			if (dhcp6_findselfsla(ifp)) {
+			if (dhcp6_findselfsla(ifp, NULL)) {
 				o.code = htons(D6_OPTION_PD_EXCLUDE);
 				memcpy(p, &o.code, sizeof(o.code));
 				p += sizeof(o.code);
@@ -1166,8 +1092,6 @@ dhcp6_sendmessage(struct interface *ifp, void (*callback)(void *))
 	uint8_t neg;
 	const char *broad_uni;
 	const struct in6_addr alldhcp = IN6ADDR_LINKLOCAL_ALLDHCP_INIT;
-	struct ipv6_addr *lla;
-	int s;
 
 	if (!callback && ifp->carrier == LINK_DOWN)
 		return 0;
@@ -1180,13 +1104,12 @@ dhcp6_sendmessage(struct interface *ifp, void (*callback)(void *))
 #endif
 
 	state = D6_STATE(ifp);
-	lla = ipv6_linklocal(ifp);
 	/* We need to ensure we have sufficient scope to unicast the address */
 	/* XXX FIXME: We should check any added addresses we have like from
 	 * a Router Advertisement */
 	if (IN6_IS_ADDR_UNSPECIFIED(&state->unicast) ||
 	    (state->state == DH6S_REQUEST &&
-	    (!IN6_IS_ADDR_LINKLOCAL(&state->unicast) || lla == NULL)))
+	    (!IN6_IS_ADDR_LINKLOCAL(&state->unicast) || !ipv6_linklocal(ifp))))
 	{
 		dst.sin6_addr = alldhcp;
 		broad_uni = "broadcasting";
@@ -1328,16 +1251,7 @@ logsend:
 		ctx->sndhdr.msg_controllen = 0;
 	}
 
-	if (ctx->dhcp6_fd != -1)
-		s = ctx->dhcp6_fd;
-	else if (lla != NULL && lla->dhcp6_fd != -1)
-		s = lla->dhcp6_fd;
-	else {
-		logerrx("%s: no socket to send from", ifp->name);
-		return -1;
-	}
-
-	if (sendmsg(s, &ctx->sndhdr, 0) == -1) {
+	if (sendmsg(ctx->dhcp6_fd, &ctx->sndhdr, 0) == -1) {
 		logerr("%s: %s: sendmsg", __func__, ifp->name);
 		/* Allow DHCPv6 to continue .... the errors
 		 * would be rate limited by the protocol.
@@ -1912,7 +1826,6 @@ dhcp6_checkstatusok(const struct interface *ifp,
 {
 	uint8_t *opt;
 	uint16_t opt_len, code;
-	size_t mlen;
 	void * (*f)(void *, size_t, uint16_t, uint16_t *), *farg;
 	char buf[32], *sbuf;
 	const char *status;
@@ -1938,8 +1851,8 @@ dhcp6_checkstatusok(const struct interface *ifp,
 
 	/* Anything after the code is a message. */
 	opt += sizeof(code);
-	mlen = opt_len - sizeof(code);
-	if (mlen == 0) {
+	opt_len = (uint16_t)(opt_len - sizeof(code));
+	if (opt_len == 0) {
 		sbuf = NULL;
 		if (code < sizeof(dhcp6_statuses) / sizeof(char *))
 			status = dhcp6_statuses[code];
@@ -1948,12 +1861,12 @@ dhcp6_checkstatusok(const struct interface *ifp,
 			status = buf;
 		}
 	} else {
-		if ((sbuf = malloc(mlen + 1)) == NULL) {
+		if ((sbuf = malloc((size_t)opt_len + 1)) == NULL) {
 			logerr(__func__);
 			return -1;
 		}
-		memcpy(sbuf, opt, mlen);
-		sbuf[mlen] = '\0';
+		memcpy(sbuf, opt, opt_len);
+		sbuf[len] = '\0';
 		status = sbuf;
 	}
 
@@ -2214,7 +2127,6 @@ dhcp6_findia(struct interface *ifp, struct dhcp6_message *m, size_t l,
 	uint8_t iaid[4];
 	char buf[sizeof(iaid) * 3];
 	struct ipv6_addr *ap;
-	struct if_ia *ifia;
 
 	if (l < sizeof(*m)) {
 		/* Should be impossible with guards at packet in
@@ -2268,9 +2180,8 @@ dhcp6_findia(struct interface *ifp, struct dhcp6_message *m, size_t l,
 		o.len = (uint16_t)(o.len - nl);
 
 		for (j = 0; j < ifo->ia_len; j++) {
-			ifia = &ifo->ia[j];
-			if (ifia->ia_type == o.code &&
-			    memcmp(ifia->iaid, ia.iaid, sizeof(ia.iaid)) == 0)
+			if (memcmp(&ifo->ia[j].iaid, ia.iaid,
+			    sizeof(ia.iaid)) == 0)
 				break;
 		}
 		if (j == ifo->ia_len &&
@@ -2282,6 +2193,13 @@ dhcp6_findia(struct interface *ifp, struct dhcp6_message *m, size_t l,
 			    buf, sizeof(buf)));
 			continue;
 		}
+		if ( j < ifo->ia_len && ifo->ia[j].ia_type != o.code) {
+			logerrx("%s: IAID %s: option type mismatch",
+			    ifp->name,
+			    hwaddr_ntoa(ia.iaid, sizeof(ia.iaid),
+			    buf, sizeof(buf)));
+			continue;
+		}
 
 		if (o.code != D6_OPTION_IA_TA) {
 			ia.t1 = ntohl(ia.t1);
@@ -2787,8 +2705,6 @@ dhcp6_delegate_prefix(struct interface *ifp)
 			}
 			for (i = 0; i < ifo->ia_len; i++) {
 				ia = &ifo->ia[i];
-				if (ia->ia_type != D6_OPTION_IA_PD)
-					continue;
 				if (memcmp(ia->iaid, ap->iaid,
 				    sizeof(ia->iaid)))
 					continue;
@@ -2885,8 +2801,6 @@ dhcp6_find_delegates(struct interface *ifp)
 				continue;
 			for (i = 0; i < ifo->ia_len; i++) {
 				ia = &ifo->ia[i];
-				if (ia->ia_type != D6_OPTION_IA_PD)
-					continue;
 				if (memcmp(ia->iaid, ap->iaid,
 				    sizeof(ia->iaid)))
 					continue;
@@ -3474,6 +3388,13 @@ dhcp6_recv(struct dhcpcd_ctx *ctx, struct ipv6_addr *ia)
 	bytes = recvmsg_realloc(s, &ctx->rcvhdr, 0);
 	if (bytes == -1) {
 		logerr("%s: recvmsg_realloc", __func__);
+		close(s);
+		eloop_event_delete(ctx->eloop, s);
+		if (ia != NULL)
+			ia->dhcp6_fd = -1;
+		else
+			ctx->dhcp6_fd = -1;
+		eloop_exit(ctx->eloop, 1);
 		return;
 	}
 	len = (size_t)bytes;
@@ -3630,16 +3551,19 @@ dhcp6_listen(struct dhcpcd_ctx *ctx, struct ipv6_addr *ia)
 	if (ia != NULL) {
 		memcpy(&sa.sin6_addr, &ia->addr, sizeof(sa.sin6_addr));
 		sa.sin6_scope_id = ia->iface->index;
-	}
+	} else if (!(ctx->options & DHCPCD_MASTER))
+		/* This socket is only used for sending. */
+		return s;
 
 	if (bind(s, (struct sockaddr *)&sa, sizeof(sa)) == -1)
 		goto errexit;
 
-	n = 1;
-	if (setsockopt(s, IPPROTO_IPV6, IPV6_RECVPKTINFO, &n, sizeof(n)) == -1)
-		goto errexit;
-
-	if (ia != NULL) {
+	if (ia == NULL) {
+		n = 1;
+		if (setsockopt(s, IPPROTO_IPV6, IPV6_RECVPKTINFO,
+		    &n, sizeof(n)) == -1)
+			goto errexit;
+	} else {
 		ia->dhcp6_fd = s;
 		eloop_event_add(ctx->eloop, s, dhcp6_recvaddr, ia);
 	}
@@ -3653,6 +3577,18 @@ errexit:
 	return -1;
 }
 
+static int
+dhcp6_open(struct dhcpcd_ctx *ctx)
+{
+
+	/* Open an unbound socket to send from. */
+	ctx->dhcp6_fd = dhcp6_listen(ctx, NULL);
+	if (ctx->dhcp6_fd != -1 && (ctx->options & DHCPCD_MASTER))
+		eloop_event_add(ctx->eloop, ctx->dhcp6_fd, dhcp6_recvctx, ctx);
+
+	return ctx->dhcp6_fd;
+}
+
 #ifndef SMALL
 static void
 dhcp6_activateinterfaces(struct interface *ifp)
@@ -3664,8 +3600,6 @@ dhcp6_activateinterfaces(struct interface *ifp)
 
 	for (i = 0; i < ifp->options->ia_len; i++) {
 		ia = &ifp->options->ia[i];
-		if (ia->ia_type != D6_OPTION_IA_PD)
-			continue;
 		for (j = 0; j < ia->sla_len; j++) {
 			sla = &ia->sla[j];
 			ifd = if_find(ifp->ctx->ifaces, sla->ifname);
@@ -3689,18 +3623,13 @@ static void
 dhcp6_start1(void *arg)
 {
 	struct interface *ifp = arg;
-	struct dhcpcd_ctx *ctx = ifp->ctx;
 	struct if_options *ifo = ifp->options;
 	struct dhcp6_state *state;
 	size_t i;
 	const struct dhcp_compat *dhc;
 
-	if (ctx->dhcp6_fd == -1 && ctx->options & DHCPCD_MASTER) {
-		ctx->dhcp6_fd = dhcp6_listen(ctx, NULL);
-		if (ctx->dhcp6_fd == -1)
-			return;
-		eloop_event_add(ctx->eloop, ctx->dhcp6_fd, dhcp6_recvctx, ctx);
-	}
+	if (ifp->ctx->dhcp6_fd == -1 && dhcp6_open(ifp->ctx) == -1)
+		return;
 
 	state = D6_STATE(ifp);
 	/* If no DHCPv6 options are configured,
@@ -3722,7 +3651,7 @@ dhcp6_start1(void *arg)
 
 #ifndef SMALL
 	/* Rapid commit won't work with Prefix Delegation Exclusion */
-	if (dhcp6_findselfsla(ifp))
+	if (dhcp6_findselfsla(ifp, NULL))
 		del_option_mask(ifo->requestmask6, D6_OPTION_RAPID_COMMIT);
 #endif
 
@@ -3929,20 +3858,22 @@ dhcp6_free(struct interface *ifp)
 void
 dhcp6_dropnondelegates(struct interface *ifp)
 {
-
 #ifndef SMALL
-	if (dhcp6_hasprefixdelegation(ifp))
-		return;
-#endif
-	if (D6_CSTATE(ifp) == NULL)
-		return;
+	struct dhcp6_state *state;
+	struct ipv6_addr *ia;
 
-	loginfox("%s: dropping DHCPv6 due to no valid routers", ifp->name);
+	if ((state = D6_STATE(ifp)) == NULL)
+		return;
+	TAILQ_FOREACH(ia, &state->addrs, next) {
+		if (ia->flags & (IPV6_AF_DELEGATED | IPV6_AF_DELEGATEDPFX))
+			return;
+	}
+#endif
 	dhcp6_drop(ifp, "EXPIRE6");
 }
 
 void
-dhcp6_handleifa(int cmd, struct ipv6_addr *ia, pid_t pid)
+dhcp6_handleifa(int cmd, struct ipv6_addr *ia)
 {
 	struct dhcp6_state *state;
 	struct interface *ifp = ia->iface;
@@ -3957,7 +3888,7 @@ dhcp6_handleifa(int cmd, struct ipv6_addr *ia, pid_t pid)
 		dhcp6_listen(ia->iface->ctx, ia);
 
 	if ((state = D6_STATE(ifp)) != NULL)
-		ipv6_handleifa_addrs(cmd, &state->addrs, ia, pid);
+		ipv6_handleifa_addrs(cmd, &state->addrs, ia);
 }
 
 ssize_t
