@@ -1,7 +1,6 @@
-/* SPDX-License-Identifier: BSD-2-Clause */
 /*
  * dhcpcd - IPv6 ND handling
- * Copyright (c) 2006-2019 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2018 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -29,8 +28,6 @@
 #ifndef IPV6ND_H
 #define IPV6ND_H
 
-#ifdef INET6
-
 #include <time.h>
 
 #include "config.h"
@@ -50,11 +47,9 @@ struct ra {
 	uint32_t reachable;
 	uint32_t retrans;
 	uint32_t mtu;
-	uint8_t hoplimit;
 	struct ipv6_addrhead addrs;
-	bool hasdns;
-	bool expired;
-	bool isreachable;
+	uint8_t hasdns;
+	uint8_t expired;
 };
 
 TAILQ_HEAD(ra_head, ra);
@@ -63,14 +58,9 @@ struct rs_state {
 	struct nd_router_solicit *rs;
 	size_t rslen;
 	int rsprobes;
-	uint32_t retrans;
-#ifdef __sun
-	int nd_fd;
-#endif
 };
 
 #define	RS_STATE(a) ((struct rs_state *)(ifp)->if_data[IF_DATA_IPV6ND])
-#define	RS_CSTATE(a) ((const struct rs_state *)(ifp)->if_data[IF_DATA_IPV6ND])
 #define	RS_STATE_RUNNING(a) (ipv6nd_hasra((a)) && ipv6nd_dadcompleted((a)))
 
 #ifndef MAX_RTR_SOLICITATION_DELAY
@@ -78,11 +68,6 @@ struct rs_state {
 #define	MAX_UNICAST_SOLICIT		3	/* 3 transmissions */
 #define	RTR_SOLICITATION_INTERVAL	4	/* seconds */
 #define	MAX_RTR_SOLICITATIONS		3	/* times */
-#define	MAX_NEIGHBOR_ADVERTISEMENT	3	/* 3 transmissions */
-
-#ifndef IPV6_DEFHLIM
-#define	IPV6_DEFHLIM			64
-#endif
 #endif
 
 /* On carrier up, expire known routers after RTR_CARRIER_EXPIRE seconds. */
@@ -96,25 +81,34 @@ struct rs_state {
 #define	RETRANS_TIMER			1000	/* milliseconds */
 #define	DELAY_FIRST_PROBE_TIME		5	/* seconds */
 
+#define	IPV6ND_REACHABLE		(1 << 0)
+#define	IPV6ND_ROUTER			(1 << 1)
+
+#ifdef INET6
 void ipv6nd_printoptions(const struct dhcpcd_ctx *,
     const struct dhcp_opt *, size_t);
 void ipv6nd_startrs(struct interface *);
-ssize_t ipv6nd_env(FILE *, const struct interface *);
+ssize_t ipv6nd_env(char **, const char *, const struct interface *);
 const struct ipv6_addr *ipv6nd_iffindaddr(const struct interface *ifp,
     const struct in6_addr *addr, unsigned int flags);
 struct ipv6_addr *ipv6nd_findaddr(struct dhcpcd_ctx *,
     const struct in6_addr *, unsigned int);
 ssize_t ipv6nd_free(struct interface *);
 void ipv6nd_expirera(void *arg);
-bool ipv6nd_hasralifetime(const struct interface *, bool);
-#define	ipv6nd_hasra(i)		ipv6nd_hasralifetime((i), false)
-bool ipv6nd_hasradhcp(const struct interface *);
-void ipv6nd_handleifa(int, struct ipv6_addr *, pid_t);
+int ipv6nd_hasra(const struct interface *);
+int ipv6nd_hasradhcp(const struct interface *);
+void ipv6nd_handleifa(int, struct ipv6_addr *);
 int ipv6nd_dadcompleted(const struct interface *);
-void ipv6nd_advertise(struct ipv6_addr *);
-void ipv6nd_startexpire(struct interface *);
+void ipv6nd_expire(struct interface *, uint32_t);
 void ipv6nd_drop(struct interface *);
-void ipv6nd_neighbour(struct dhcpcd_ctx *, struct in6_addr *, bool);
-#endif /* INET6 */
+void ipv6nd_neighbour(struct dhcpcd_ctx *, struct in6_addr *, int);
+#else
+#define ipv6nd_startrs(a) {}
+#define ipv6nd_free(a) {}
+#define ipv6nd_hasra(a) (0)
+#define ipv6nd_dadcompleted(a) (0)
+#define ipv6nd_drop(a) {}
+#define ipv6nd_expire(a, b) {}
+#endif
 
-#endif /* IPV6ND_H */
+#endif
